# Copyright 2006, 2007, 2008, 2009 Brailcom, o.p.s.
#
# Author: Tomas Cerha <cerha@brailcom.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=broad-exception-caught
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals

"""Provides an Orca speech server for Speech Dispatcher backend."""

from __future__ import annotations

import gc
import locale
import time
from typing import TYPE_CHECKING, Any

from gi.repository import GLib

from . import debug, focus_manager, guilabels, mathsymbols, speechserver
from .acss import ACSS
from .ax_utilities import AXUtilities
from .speechserver import CapitalizationStyle, PunctuationStyle
from .ssml import SSML, SSMLCapabilities

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import ClassVar

    from . import input_event
    from .speechserver import VoiceFamily

try:
    import speechd
except Exception:
    _SPEECHD_AVAILABLE = False
else:
    _SPEECHD_AVAILABLE = True
    try:
        _CallbackType = speechd.CallbackType
    except AttributeError:
        _SPEECHD_VERSION_OK = False
    else:
        _SPEECHD_VERSION_OK = True


class SpeechServer(speechserver.SpeechServer):
    """Speech Dispatcher speech server for Orca."""

    _active_servers: ClassVar[dict[str, SpeechServer]] = {}

    DEFAULT_SERVER_ID = "default"
    _SERVER_NAMES: ClassVar[dict[str, str]] = {DEFAULT_SERVER_ID: guilabels.DEFAULT_SYNTHESIZER}

    @staticmethod
    def get_factory_name() -> str:
        return guilabels.SPEECH_DISPATCHER

    @staticmethod
    def get_speech_servers() -> list[SpeechServer]:
        servers = []
        default = SpeechServer._get_speech_server(SpeechServer.DEFAULT_SERVER_ID)
        if default is not None:
            servers.append(default)
            for module in default.list_output_modules():
                server = SpeechServer._get_speech_server(module)
                if server is not None:
                    servers.append(server)
        return servers

    @classmethod
    def _get_speech_server(cls, server_id: str) -> SpeechServer | None:
        """Return an active server for given id.

        Attempt to create the server if it doesn't exist yet.  Returns None
        when it is not possible to create the server.
        """

        if server_id not in cls._active_servers:
            cls(server_id)
        # Don't return the instance, unless it is successfully added
        # to `_active_Servers'.
        return cls._active_servers.get(server_id)

    @staticmethod
    def get_speech_server(info: list[str] | None = None) -> SpeechServer | None:
        """Gets a given SpeechServer based upon the info."""

        this_id = info[1] if info is not None else SpeechServer.DEFAULT_SERVER_ID
        return SpeechServer._get_speech_server(this_id)

    @staticmethod
    def shutdown_active_servers() -> None:
        servers = list(SpeechServer._active_servers.values())
        for server in servers:
            server.shutdown()

    # *** Instance methods ***

    def __init__(self, server_id: str) -> None:
        super().__init__()
        self._id = server_id
        self._client: Any = None
        self._output_module: str | None = None
        self._default_voice: dict[str, Any] = {}
        self._current_voice_properties: dict[str, Any] = {}
        self._current_synthesis_voice: str | None = None
        self._current_punctuation_level: PunctuationStyle = PunctuationStyle.MOST
        self._current_capitalization_style: str = CapitalizationStyle.NONE.value
        self._voice_families_cache: dict[
            tuple[str, str, str | None, int | None],
            list[tuple[str, str, str | None]],
        ] = {}
        self._acss_manipulators = (
            (ACSS.FAMILY, self._set_family),
            (ACSS.RATE, self._set_rate),
            (ACSS.AVERAGE_PITCH, self._set_pitch),
            (ACSS.GAIN, self._set_volume),
        )
        if not _SPEECHD_AVAILABLE:
            msg = "ERROR: Speech Dispatcher is not available"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return
        if not _SPEECHD_VERSION_OK:
            msg = "ERROR: Speech Dispatcher version 0.6.2 or later is required."
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return
        # The following constants must be initialized in runtime since they
        # depend on the speechd module being available.
        try:
            most = speechd.PunctuationMode.MOST
        except Exception:
            most = speechd.PunctuationMode.SOME
        self._punctuation_mode_map = {
            PunctuationStyle.ALL: speechd.PunctuationMode.ALL,
            PunctuationStyle.MOST: most,
            PunctuationStyle.SOME: speechd.PunctuationMode.SOME,
            PunctuationStyle.NONE: speechd.PunctuationMode.NONE,
        }
        self._callback_type_map = {
            speechd.CallbackType.BEGIN: speechserver.SayAllContext.PROGRESS,
            speechd.CallbackType.CANCEL: speechserver.SayAllContext.INTERRUPTED,
            speechd.CallbackType.END: speechserver.SayAllContext.COMPLETED,
            speechd.CallbackType.INDEX_MARK: speechserver.SayAllContext.PROGRESS,
        }

        self._default_voice_name = guilabels.SPEECH_DEFAULT_VOICE % server_id

        try:
            self._init()
        except Exception:
            debug.print_exception(debug.LEVEL_WARNING)
            msg = "ERROR: Speech Dispatcher service failed to connect"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
        else:
            SpeechServer._active_servers[server_id] = self

    def _init(self) -> None:
        try:
            self._client = client = speechd.SSIPClient("Orca", component=self._id)
        except (speechd.SSIPCommunicationError, speechd.SpawnError) as error:
            msg = f"ERROR: Failed to connect to Speech Dispatcher: {error}"
            debug.print_message(debug.LEVEL_SEVERE, msg, True)
            self._client = None
            return
        client.set_priority(speechd.Priority.MESSAGE)

        # The speechServerInfo setting is not connected to the speechServerFactory. As a result,
        # the user's chosen server (synthesizer) might be from spiel. Don't mutate self._id
        # here because it must stay consistent with the _active_servers key for proper cleanup.
        # _output_module (set via set_output_module) takes precedence because it reflects
        # what was explicitly configured; fall back to self._id for non-default servers.
        module = self._output_module
        if not module and self._id and self._id != self.DEFAULT_SERVER_ID:
            module = self._id
        if module:
            try:
                available_modules = client.list_output_modules()
                if module in available_modules:
                    client.set_output_module(module)
            except (AttributeError, speechd.SSIPCommandError):
                pass
        self._current_voice_properties = {}
        mode = self._punctuation_mode_map[self._current_punctuation_level]
        client.set_punctuation(mode)
        client.set_data_mode(speechd.DataMode.SSML)

    def update_capitalization_style(self, style: str) -> None:
        """Updates the capitalization style used by the speech server."""

        self._current_capitalization_style = style
        if self._client is None:
            return

        # "spell" is only applied per-character in speak_character() to
        # prevent the synthesizer from announcing capitals in words and phrases.
        if style == CapitalizationStyle.ICON.value:
            sd_style = "icon"
        else:
            sd_style = "none"

        try:
            self._client.set_cap_let_recogn(sd_style)
        except speechd.SSIPCommunicationError:
            msg = "SPEECH DISPATCHER: Connection lost. Trying to reconnect."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.reset()
            if self._client is not None:
                self._client.set_cap_let_recogn(sd_style)
        except Exception:
            pass

    def update_punctuation_level(self, level: PunctuationStyle) -> None:
        """Punctuation level changed, inform this speechServer."""
        self._current_punctuation_level = level
        if self._client is None:
            return
        mode = self._punctuation_mode_map[level]
        self._client.set_punctuation(mode)

    def _send_command(self, command: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        try:
            return command(*args, **kwargs)
        except speechd.SSIPCommunicationError:
            msg = "SPEECH DISPATCHER: Connection lost. Trying to reconnect."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.reset()
            if self._client is None:
                return None
            try:
                return command(*args, **kwargs)
            except (AttributeError, speechd.SSIPCommandError) as error:
                msg = f"SPEECH DISPATCHER: Failed to reconnect: {error}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)
                return None
        except Exception:
            return None

    def _set_rate(self, acss_rate: float) -> None:
        if self._client is None:
            return
        rate = int(2 * max(0, min(99, acss_rate)) - 98)
        self._send_command(self._client.set_rate, rate)

    def _set_pitch(self, acss_pitch: float) -> None:
        if self._client is None:
            return
        pitch = int(20 * max(0, min(9, acss_pitch)) - 90)
        self._send_command(self._client.set_pitch, pitch)

    def _set_volume(self, acss_volume: float) -> None:
        if self._client is None:
            return
        volume = int(15 * max(0, min(9, acss_volume)) - 35)
        self._send_command(self._client.set_volume, volume)

    def _get_language_and_dialect(self, acss_family: dict[str, Any] | None) -> tuple[str, str]:
        if acss_family is None:
            acss_family = {}

        language = acss_family.get(speechserver.VoiceFamily.LANG)
        dialect = acss_family.get(speechserver.VoiceFamily.DIALECT)

        if not language:
            family_locale, _encoding = locale.getlocale()

            language, dialect = "", ""
            if family_locale:
                locale_values = family_locale.split("_")
                language = locale_values[0]
                if len(locale_values) == 2:
                    dialect = locale_values[1]

        return str(language), str(dialect)

    def _set_family(self, acss_family: dict[str, Any] | None) -> None:
        if self._client is None:
            return

        lang, dialect = self._get_language_and_dialect(acss_family)
        if lang:
            self._send_command(self._client.set_language, lang)
            if dialect:
                # Try to set precise dialect
                self._send_command(self._client.set_language, lang + "-" + dialect)

        if acss_family is not None:
            if name := acss_family.get(speechserver.VoiceFamily.NAME):
                self._send_command(self._client.set_synthesis_voice, name)
                self._current_synthesis_voice = name

    def set_default_voice(self, default_voice: dict[str, Any]) -> None:
        """Sets the default voice ACSS properties for fallback use."""

        self._default_voice = default_voice

    def _debug_sd_values(self, prefix: str = "") -> None:
        if debug.debugLevel > debug.LEVEL_INFO:
            return

        try:
            if self._client is not None:
                sd_rate = self._send_command(self._client.get_rate)
                sd_pitch = self._send_command(self._client.get_pitch)
                sd_volume = self._send_command(self._client.get_volume)
                sd_language = self._send_command(self._client.get_language)
            else:
                sd_rate = sd_pitch = sd_volume = sd_language = "(client not available)"
        except Exception:
            sd_rate = sd_pitch = sd_volume = sd_language = "(exception occurred)"

        family = self._current_voice_properties.get(ACSS.FAMILY) or {}

        punctuation_style = self._current_punctuation_level.name

        msg = (
            f"SPEECH DISPATCHER: {prefix}\n"
            f"ORCA rate {self._current_voice_properties.get(ACSS.RATE)}, "
            f"pitch {self._current_voice_properties.get(ACSS.AVERAGE_PITCH)}, "
            f"volume {self._current_voice_properties.get(ACSS.GAIN)}, "
            f"language {self._get_language_and_dialect(family)[0]}, "
            f"punctuation: {punctuation_style}\n"
            f"SD rate {sd_rate}, pitch {sd_pitch}, volume {sd_volume}, language {sd_language}"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _apply_acss(self, acss: dict[str, Any] | None) -> None:
        merged: dict[str, Any] = dict(self._default_voice)
        if acss is not None:
            merged.update(acss)
        current = self._current_voice_properties
        for acss_property, method in self._acss_manipulators:
            value = merged.get(acss_property)
            if value is not None:
                if current.get(acss_property) != value:
                    method(value)
                    current[acss_property] = value
            elif acss_property == ACSS.AVERAGE_PITCH:
                self._set_pitch(5.0)
                current[acss_property] = 5.0
            elif acss_property == ACSS.GAIN:
                self._set_volume(10.0)
                current[acss_property] = 10.0
            elif acss_property == ACSS.RATE:
                self._set_rate(50.0)
                current[acss_property] = 50.0
            elif acss_property == ACSS.FAMILY:
                self._set_family({})
                current[acss_property] = {}

    def _speak(self, text: str, acss: dict[str, Any] | None, **kwargs: Any) -> None:
        if isinstance(text, ACSS):
            text = ""

        ssml = SSML.markup_text(text, SSMLCapabilities.MARK)

        self._apply_acss(acss)
        self._debug_sd_values(f"Speaking '{ssml}' ")
        if self._client is not None:
            self._send_command(self._client.speak, ssml, **kwargs)

    def _say_all(
        self,
        iterator: Iterator[tuple[speechserver.SayAllContext, dict[str, Any]]],
        orca_callback: Callable[[speechserver.SayAllContext, int], None],
    ) -> bool:
        """Process another sayAll chunk.

        Called by the gidle thread.

        """
        try:
            context, acss = next(iterator)
        except StopIteration:
            pass
        else:

            def callback(callback_type, index_mark=None):
                # This callback is called in Speech Dispatcher listener thread.
                # No subsequent Speech Dispatcher interaction is allowed here,
                # so we pass the calls to the gidle thread.
                t = self._callback_type_map[callback_type]
                if t == speechserver.SayAllContext.PROGRESS:
                    if index_mark:
                        index = index_mark.split(":")
                        if len(index) >= 2:
                            start, end = index[0:2]
                            context.current_offset = context.start_offset + int(start)
                            context.current_end_offset = context.start_offset + int(end)
                            msg = (
                                f"SPEECH DISPATCHER: Got mark "
                                f"{context.current_offset}:{context.current_end_offset} / "
                                f"{context.start_offset}:{context.end_offset}"
                            )
                            debug.print_message(debug.LEVEL_INFO, msg, True)
                    else:
                        context.current_offset = context.start_offset
                        context.current_end_offset = None
                elif t == speechserver.SayAllContext.COMPLETED:
                    context.current_offset = context.end_offset
                    context.current_end_offset = None
                GLib.idle_add(orca_callback, context.copy(), t)
                if t == speechserver.SayAllContext.COMPLETED:
                    GLib.idle_add(self._say_all, iterator, orca_callback)

            self._speak(
                context.utterance,
                acss,
                callback=callback,
                event_types=list(self._callback_type_map.keys()),
            )
        return False  # to indicate, that we don't want to be called again.

    def _cancel(self) -> None:
        if self._client is not None:
            self._send_command(self._client.cancel)

    def _change_default_speech_rate(self, step: int, decrease: bool = False) -> None:
        delta = step * (-1 if decrease else 1)
        rate = self._default_voice.get(ACSS.RATE, 50)
        new_rate = max(0, min(99, rate + delta))
        self._default_voice[ACSS.RATE] = new_rate
        self._set_rate(new_rate)
        self._current_voice_properties[ACSS.RATE] = new_rate
        msg = f"SPEECH DISPATCHER: Rate set to {new_rate}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _change_default_speech_pitch(self, step: float, decrease: bool = False) -> None:
        delta = step * (-1 if decrease else 1)
        pitch = self._default_voice.get(ACSS.AVERAGE_PITCH, 5)
        new_pitch = max(0, min(9, pitch + delta))
        self._default_voice[ACSS.AVERAGE_PITCH] = new_pitch
        self._set_pitch(new_pitch)
        self._current_voice_properties[ACSS.AVERAGE_PITCH] = new_pitch
        msg = f"SPEECH DISPATCHER: Pitch set to {new_pitch}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _change_default_speech_volume(self, step: float, decrease: bool = False) -> None:
        delta = step * (-1 if decrease else 1)
        volume = self._default_voice.get(ACSS.GAIN, 10)
        new_volume = max(0, min(9, volume + delta))
        self._default_voice[ACSS.GAIN] = new_volume
        self._set_volume(new_volume)
        self._current_voice_properties[ACSS.GAIN] = new_volume
        msg = f"SPEECH DISPATCHER: Volume set to {new_volume}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def get_info(self) -> list[str]:
        return [self._SERVER_NAMES.get(self._id, self._id), self._id]

    def _get_default_voice_language(
        self,
        voices: tuple[tuple[str, str, str | None], ...],
    ) -> str:
        """Returns the default language string based on the current locale and available voices."""

        current_locale = locale.getlocale(locale.LC_MESSAGES)[0]
        if current_locale is None or "_" not in current_locale:
            return ""

        locale_lang, locale_dialect = current_locale.split("_")
        locale_language = locale_lang + "-" + locale_dialect
        for _name, lang, _variant in voices:
            if lang == locale_language:
                return locale_language
        for _name, lang, _variant in voices:
            if lang == locale_lang:
                return locale_lang
        return locale_language

    def get_voice_families(self) -> list[speechserver.VoiceFamily]:
        """Returns the list of voice families available in the current synthesizer."""

        voices: tuple[tuple[str, str, str | None], ...] = ()
        if self._client is not None:
            try:
                # This command is not available with older SD versions.
                list_synthesis_voices = self._client.list_synthesis_voices
            except AttributeError:
                pass
            else:
                try:
                    result = self._send_command(list_synthesis_voices)
                    if result:
                        voices = tuple(result)
                except Exception:
                    pass

        default_lang = self._get_default_voice_language(voices)
        voices = ((self._default_voice_name, default_lang, None), *voices)

        families = []
        for name, lang, variant in voices:
            families.append(
                speechserver.VoiceFamily(
                    {
                        speechserver.VoiceFamily.NAME: name,
                        # speechserver.VoiceFamily.GENDER: speechserver.VoiceFamily.MALE,
                        speechserver.VoiceFamily.LANG: lang.partition("-")[0],
                        speechserver.VoiceFamily.DIALECT: lang.partition("-")[2],
                        speechserver.VoiceFamily.VARIANT: variant,
                    },
                ),
            )

        return families

    def speak(self, text: str | None = None, acss: dict[str, Any] | None = None) -> None:
        if not text:
            return

        if len(text) == 1:
            if (
                text.isupper()
                and text.isalpha()
                and self._current_capitalization_style == CapitalizationStyle.ICON.value
            ):
                # See https://gitlab.gnome.org/GNOME/orca/-/issues/657
                self._apply_acss(acss)
                self._speak(text, acss)
                return
            msg = f"SPEECH DISPATCHER: Speaking '{text}' as char"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._apply_acss(acss)
            if self._client is not None:
                self._send_command(self._client.char, text)
        else:
            msg = f"SPEECH DISPATCHER: Speaking '{text}' as string"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._speak(text, acss)

    def say_all(
        self,
        utterance_iterator: Iterator[tuple[speechserver.SayAllContext, dict[str, Any]]],
        progress_callback: Callable[[speechserver.SayAllContext, int], None],
    ) -> None:
        """Iterates through the given utterance_iterator, speaking each utterance."""

        GLib.idle_add(self._say_all, utterance_iterator, progress_callback)

    def speak_character(
        self,
        character: str,
        acss: dict[str, Any] | None = None,
        cap_style: CapitalizationStyle | None = None,
    ) -> None:
        """Speaks character."""

        self._apply_acss(acss)
        if cap_style == CapitalizationStyle.NONE:
            character = character.lower()

        name = character
        focus = focus_manager.get_manager().get_locus_of_focus()
        if AXUtilities.is_math_related(focus):
            # TODO - JD: If we're reaching this point without the name having been adjusted, we're
            # doing it wrong.
            name = mathsymbols.get_character_name(character)

        if not name or name == character:
            if cap_style == CapitalizationStyle.ICON:
                # char() does not reliably trigger the icon sound whereas
                # speak() does. See https://gitlab.gnome.org/GNOME/orca/-/issues/657
                self._speak(character, acss)
                return
            msg = f"SPEECH DISPATCHER: Speaking '{character}' as char"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if self._client is not None:
                if cap_style == CapitalizationStyle.SPELL:
                    self._send_command(self._client.set_cap_let_recogn, "spell")
                self._send_command(self._client.char, character)
                if cap_style == CapitalizationStyle.SPELL:
                    self._send_command(self._client.set_cap_let_recogn, "none")
            return

        self.speak(name, acss)

    def speak_key_event(
        self,
        event: input_event.KeyboardEvent,
        acss: dict[str, Any] | None = None,
    ) -> None:
        """Speaks event."""

        event_string = event.get_key_name()
        locking_state_string = event.get_locking_state_string()
        event_string = f"{event_string} {locking_state_string}".strip()
        # Speaking this key event as a character makes the user experience more consistent.
        # Example: The user presses '.' and then presses BackSpace and expects the output to
        # be the same.
        if len(event_string) == 1:
            msg = f"SPEECH DISPATCHER: Speaking '{event_string}' as char"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._apply_acss(acss)
            if self._client is not None:
                self._send_command(self._client.char, event_string)
        else:
            msg = f"SPEECH DISPATCHER: Speaking '{event_string}' as string"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.speak(event_string, acss=acss)

    def increase_speech_rate(self, step: int = 5) -> None:
        self._change_default_speech_rate(step)

    def decrease_speech_rate(self, step: int = 5) -> None:
        self._change_default_speech_rate(step, decrease=True)

    def increase_speech_pitch(self, step: float = 0.5) -> None:
        self._change_default_speech_pitch(step)

    def decrease_speech_pitch(self, step: float = 0.5) -> None:
        self._change_default_speech_pitch(step, decrease=True)

    def increase_speech_volume(self, step: float = 0.5) -> None:
        self._change_default_speech_volume(step)

    def decrease_speech_volume(self, step: float = 0.5) -> None:
        self._change_default_speech_volume(step, decrease=True)

    def _normalized_language_and_dialect(self, language: str, dialect: str = "") -> tuple[str, str]:
        """Attempts to ensure consistency across inconsistent formats."""

        if "-" in language:
            normalized_language = language.split("-", 1)[0].lower()
            normalized_dialect = language.split("-", 1)[-1].lower()
        else:
            normalized_language = language.lower()
            normalized_dialect = dialect.lower()

        return normalized_language, normalized_dialect

    def clear_voice_families_cache(self) -> None:
        """Clear the cache for voice family lookups."""

        msg = "SPEECH DISPATCHER: Clearing voice families cache"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._voice_families_cache.clear()

    def clear_cached_voice_properties(self) -> None:
        """Clear cached voice properties to force reapplication on next speech."""

        msg = "SPEECH DISPATCHER: Clearing cached voice properties"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._current_voice_properties.clear()

    def _filter_voices_for_language(
        self,
        voices: list[tuple[str, str, str | None]],
        target_language: str,
        target_dialect: str,
        maximum: int | None,
    ) -> tuple[list[tuple[str, str, str | None]], list[tuple[str, str, str | None]]]:
        """Filters voices by language and dialect, returning candidates and fallbacks."""

        candidates: list[tuple[str, str, str | None]] = []
        fallbacks: list[tuple[str, str, str | None]] = []
        for voice in voices:
            normalized_language, normalized_dialect = self._normalized_language_and_dialect(
                voice[1],
            )
            if normalized_language != target_language:
                continue
            if normalized_dialect == target_dialect or (
                not normalized_dialect and target_dialect == normalized_language
            ):
                candidates.append(voice)
            elif not target_dialect:
                if normalized_dialect == target_language:
                    candidates.append(voice)
                if len(normalized_dialect) == 2:
                    fallbacks.append(voice)
            if maximum is not None and len(candidates) >= maximum:
                break
        return candidates, fallbacks

    def get_voice_families_for_language(
        self,
        language: str,
        dialect: str = "",
        variant: str | None = None,
        maximum: int | None = None,
    ) -> list[tuple[str, str, str | None]]:
        """Returns the families for language available in the current synthesizer."""

        if not language:
            language, dialect = self._get_language_and_dialect(None)
        elif not dialect:
            locale_language, locale_dialect = self._get_language_and_dialect(None)
            if language == locale_language:
                dialect = locale_dialect

        cache_key = (language, dialect, variant, maximum)
        if cache_key in self._voice_families_cache:
            msg = (
                f"SPEECH DISPATCHER: Returning cached result for language='{language}' "
                f"dialect='{dialect}' variant='{variant}'"
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return self._voice_families_cache[cache_key]
        start = time.time()
        target_language, target_dialect = self._normalized_language_and_dialect(language, dialect)

        result = []
        if self._client is None:
            return result

        msg = (
            f"SPEECH DISPATCHER: Searching for language='{language}' "
            f"dialect='{dialect}' variant='{variant}'."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

        language_with_dialect = f"{language}-{dialect}" if dialect else language
        try:
            voices = self._client.list_synthesis_voices(language_with_dialect, variant)
            msg = f"SPEECH DISPATCHER: Unfiltered voice list has {len(voices)} entries."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if not voices:
                voices = self._client.list_synthesis_voices(language, variant)
                msg = f"SPEECH DISPATCHER: Unfiltered voice list (try 2) has {len(voices)} entries."
                debug.print_message(debug.LEVEL_INFO, msg, True)
        except (AttributeError, ValueError) as error:
            msg = f"SPEECH DISPATCHER: specifying language and variant failed: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            try:
                voices = self._client.list_synthesis_voices()
            except (AttributeError, ValueError, speechd.SSIPCommandError) as error2:
                msg = f"SPEECH DISPATCHER: Error listing synthesis voices: {error2}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)
                return []
        except speechd.SSIPCommandError as error:
            msg = f"SPEECH DISPATCHER: Error listing synthesis voices: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return []

        candidates, fallbacks = self._filter_voices_for_language(
            voices,
            target_language,
            target_dialect,
            maximum,
        )

        msg = (
            f"SPEECH DISPATCHER: Found {len(candidates)} match(es) for "
            f"language='{language}' dialect='{dialect}' variant='{variant}' "
            f"in {time.time() - start:.4f}s."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        if not candidates and fallbacks:
            msg = f"SPEECH DISPATCHER: No direct matches found, using {len(fallbacks)} fallback(s)."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            candidates = fallbacks

        self._voice_families_cache[cache_key] = candidates
        return candidates

    def get_output_module(self) -> str:
        if self._client is None:
            return ""
        result = self._send_command(self._client.get_output_module)
        if result is not None:
            return result
        return ""

    def set_output_module(self, module_id: str) -> None:
        """Set the speech output module to the specified provider."""

        # TODO - JD: This updates the output module, but not the the value of self._id.
        # That might be desired (e.g. self._id impacts what is shown in Orca preferences),
        # but it can be confusing.
        self._output_module = module_id
        if self._client is not None:
            self._send_command(self._client.set_output_module, module_id)
            self.clear_voice_families_cache()
            # Update the default voice name to match the new module
            self._default_voice_name = guilabels.SPEECH_DEFAULT_VOICE % module_id

    def stop(self) -> None:
        self._cancel()

    def shutdown(self) -> None:
        try:
            # Don't call _cancel() here because it can cut off messages we want to complete, such
            # as "screen reader off."
            if self._client is not None:
                self._client.close()
                # Set client to None to allow immediate garbage collection.
                self._client = None
                # Force garbage collection to clean up Speech Dispatcher objects immediately
                # This prevents hanging during Python's final garbage collection.
                gc.collect()
        except Exception as error:
            msg = f"SPEECH DISPATCHER: Error during shutdown of server {self._id}: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
        finally:
            if self._id in SpeechServer._active_servers:
                del SpeechServer._active_servers[self._id]

    def reset(self) -> None:
        try:
            msg = f"SPEECH DISPATCHER: Resetting server {self._id}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if self._client is not None:
                self._client.close()
        except Exception as error:
            msg = f"SPEECH DISPATCHER: Error during reset of server {self._id}: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
        finally:
            self.clear_voice_families_cache()
            self._init()

    def list_output_modules(self) -> tuple[str, ...]:
        """Return names of available output modules as a tuple of strings."""

        if self._client is None:
            return ()
        try:
            return self._send_command(self._client.list_output_modules)
        except AttributeError:
            return ()
        except speechd.SSIPCommandError:
            return ()

    def get_voice_family(self) -> VoiceFamily:
        """Returns the current voice family as a VoiceFamily dictionary."""

        voice_name = self._current_synthesis_voice or self._default_voice_name
        language = ""
        try:
            language = self._send_command(self._client.get_language)
        except (AttributeError, speechd.SSIPCommandError) as error:
            msg = f"SPEECH DISPATCHER: Error getting language: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)

        lang_parts = language.partition("-")
        lang = lang_parts[0]
        dialect = lang_parts[2]
        return speechserver.VoiceFamily(
            {
                speechserver.VoiceFamily.NAME: voice_name,
                speechserver.VoiceFamily.LANG: lang,
                speechserver.VoiceFamily.DIALECT: dialect,
                speechserver.VoiceFamily.VARIANT: None,
            },
        )

    def set_voice_family(self, family: VoiceFamily) -> None:
        """Sets the voice family to family VoiceFamily dictionary."""

        if not family:
            return

        voice_name = family.get(speechserver.VoiceFamily.NAME, "")
        language = family.get(speechserver.VoiceFamily.LANG, "")
        dialect = family.get(speechserver.VoiceFamily.DIALECT, "")
        if dialect:
            language = f"{language}-{dialect}" if language else dialect

        # Try to set synthesis voice (individual voice like Nathan, Federica)
        if voice_name:
            try:
                self._send_command(self._client.set_synthesis_voice, voice_name)
                self._current_synthesis_voice = voice_name
            except AttributeError as error:
                msg = f"SPEECH DISPATCHER: Synthesis voice not supported: {error}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self._current_synthesis_voice = None
            except speechd.SSIPCommandError as error:
                msg = f"SPEECH DISPATCHER: Error setting synthesis voice {voice_name}: {error}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)

        if language:
            try:
                self._send_command(self._client.set_language, language)
            except speechd.SSIPCommandError as error:
                msg = f"SPEECH DISPATCHER: Error setting language {language}: {error}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Shared migration logic for converting between JSON settings and GSettings.

This module has no imports of debug, settings_manager, or anything that pulls
in Atspi, so it can be used by both the orca runtime and the standalone
gsettings_import_export tool.
"""

from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any

from gi.repository import Gio, GLib

VOICE_TYPES: list[str] = ["default", "uppercase", "hyperlink", "system"]

# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
VOICE_MIGRATION_MAP: dict[str, tuple[str, str, Any]] = {
    "established": ("established", "b", False),
    "rate": ("rate", "i", 50),
    "average-pitch": ("pitch", "d", 5.0),
    "gain": ("volume", "d", 10.0),
}

# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
VOICE_FAMILY_FIELDS: dict[str, str] = {
    "name": "family-name",
    "lang": "family-lang",
    "dialect": "family-dialect",
    "gender": "family-gender",
    "variant": "family-variant",
}

# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
LEGACY_KEY_ALIASES: dict[str, str] = {
    "progressBarVerbosity": "progressBarSpeechVerbosity",
    "progressBarUpdateInterval": "progressBarSpeechInterval",
}

# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
REVERSE_LEGACY_KEY_ALIASES: dict[str, str] = {v: k for k, v in LEGACY_KEY_ALIASES.items()}

# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
KEYBINDINGS_METADATA_KEYS: frozenset[str] = frozenset({"keyboardLayout", "orcaModifierKeys"})


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
@dataclass
class SettingsMapping:
    """Describes a mapping between a preferences key and a GSettings key."""

    migration_key: str
    gs_key: str
    gtype: str  # "b", "s", "i", "d", "as"
    default: Any
    enum_map: dict[int, str] | None = None
    string_enum: bool = False


def sanitize_gsettings_path(name: str) -> str:
    """Sanitize a name for use in a GSettings path."""

    sanitized = name.lower()
    sanitized = re.sub(r"[^a-z0-9-]", "-", sanitized)
    sanitized = re.sub(r"-+", "-", sanitized)
    return sanitized.strip("-")


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def resolve_enum_nick(value: Any, enum_map: dict[int, str]) -> str | None:
    """Resolve a JSON enum value (string nick, int, or bool) to a GSettings enum nick."""

    if isinstance(value, str):
        return value if value in enum_map.values() else None
    return enum_map.get(value)


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
_NAVIGATION_ENABLED_KEYS = (
    "caretNavigationEnabled",
    "structuralNavigationEnabled",
    "tableNavigationEnabled",
)


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def force_navigation_enabled(json_dict: dict) -> None:
    """Force navigation-enabled keys to True during migration."""

    # In the old system these defaulted to False because per-script logic controlled
    # activation. In the new system they represent the user's preference and default
    # to True. Migrating the old False would disable navigation for all scripts.
    for key in _NAVIGATION_ENABLED_KEYS:
        if key in json_dict:
            json_dict[key] = True


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def fix_bool_enum_values(json_dict: dict) -> None:
    """Fix enum settings stored as booleans instead of integers in JSON."""

    # Some enum-typed settings are booleans in user JSON files instead of the
    # expected integers. Python's bool-is-int means False==0 and True==1,
    # which silently resolve to the wrong enum nick during migration.
    bool_enum_fixes: dict[str, dict[bool, int]] = {
        "findResultsVerbosity": {True: 2, False: 2},  # FIND_SPEAK_ALL
    }
    for key, mapping in bool_enum_fixes.items():
        if key in json_dict and isinstance(json_dict[key], bool):
            json_dict[key] = mapping[json_dict[key]]


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def apply_legacy_aliases(json_dict: dict) -> None:
    """Copy legacy key names to their modern equivalents if modern key is absent."""

    for legacy_key, modern_key in LEGACY_KEY_ALIASES.items():
        if legacy_key in json_dict and modern_key not in json_dict:
            json_dict[modern_key] = json_dict[legacy_key]


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def add_legacy_aliases(result: dict) -> None:
    """Add legacy key names to exported dict for backwards compatibility with stable."""

    for modern_key, legacy_key in REVERSE_LEGACY_KEY_ALIASES.items():
        if modern_key in result:
            result[legacy_key] = result[modern_key]


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def hoist_keybindings_metadata(profile_prefs: dict) -> None:
    """Move keyboardLayout and orcaModifierKeys from keybindings to profile level."""

    keybindings = profile_prefs.get("keybindings", {})
    for key in KEYBINDINGS_METADATA_KEYS:
        if key in keybindings and key not in profile_prefs:
            profile_prefs[key] = keybindings[key]


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def sink_keybindings_metadata(profile_data: dict, keybindings: dict) -> None:
    """Move keyboardLayout and orcaModifierKeys from profile level into keybindings."""

    for key in KEYBINDINGS_METADATA_KEYS:
        if key in profile_data:
            keybindings[key] = profile_data.pop(key)


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
DESKTOP_MODIFIER_KEYS_DEFAULT: list[str] = ["Insert", "KP_Insert"]
LAPTOP_MODIFIER_KEYS_DEFAULT: list[str] = ["Caps_Lock", "Shift_Lock"]


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def populate_per_layout_modifier_keys(
    gs: Gio.Settings,
    profile_prefs: dict,
    skip_defaults: bool,
) -> bool:
    """Populates keyboard-layout and modifier keys from keybinding metadata during migration."""

    layout = profile_prefs.get("keyboardLayout", 1)
    is_desktop = layout == 1
    wrote_any = False

    layout_nick = "desktop" if is_desktop else "laptop"
    if not skip_defaults or layout_nick != "desktop":
        gs.set_string("keyboard-layout", layout_nick)
        wrote_any = True

    modifier_keys = profile_prefs.get("orcaModifierKeys")
    if modifier_keys is None:
        return wrote_any

    if is_desktop:
        key = "desktop-modifier-keys"
        default = DESKTOP_MODIFIER_KEYS_DEFAULT
    else:
        key = "laptop-modifier-keys"
        default = LAPTOP_MODIFIER_KEYS_DEFAULT

    if skip_defaults and modifier_keys == default:
        return wrote_any

    gs.set_strv(key, modifier_keys)
    return True


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def _write_one_mapping(
    gs: Gio.Settings,
    m: SettingsMapping,
    value: Any,
    skip_defaults: bool,
) -> bool:
    """Write a single mapping value to GSettings. Returns True if written."""

    if m.enum_map is not None:
        nick = resolve_enum_nick(value, m.enum_map)
        if nick is None:
            return False
        if skip_defaults and m.default in (nick, value):
            return False
        gs.set_string(m.gs_key, nick)
        return True

    if skip_defaults and value == m.default:
        return False
    if m.gtype == "b":
        gs.set_boolean(m.gs_key, value)
    elif m.gtype == "s":
        gs.set_string(m.gs_key, value)
    elif m.gtype == "i":
        gs.set_int(m.gs_key, int(value))
    elif m.gtype == "d":
        gs.set_double(m.gs_key, float(value))
    elif m.gtype == "as":
        gs.set_strv(m.gs_key, value)
    else:
        return False
    return True


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def json_to_gsettings(
    json_dict: dict,
    gs: Gio.Settings,
    mappings: list[SettingsMapping],
    skip_defaults: bool = True,
) -> bool:
    """Writes JSON settings to a Gio.Settings object. Returns True if any value was written."""

    wrote_any = False
    for m in mappings:
        if m.migration_key not in json_dict:
            continue
        if _write_one_mapping(gs, m, json_dict[m.migration_key], skip_defaults):
            wrote_any = True
    return wrote_any


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def gsettings_to_json(gs: Gio.Settings, mappings: list[SettingsMapping]) -> dict:
    """Reads explicitly-set GSettings values into a JSON-compatible dict."""

    result: dict[str, Any] = {}
    for m in mappings:
        user_value = gs.get_user_value(m.gs_key)
        if user_value is None:
            continue
        if m.enum_map is not None:
            gs_str = user_value.get_string()
            if m.string_enum:
                result[m.migration_key] = gs_str
            else:
                reverse_map = {v: k for k, v in m.enum_map.items()}
                json_value = reverse_map.get(gs_str)
                if json_value is not None:
                    result[m.migration_key] = json_value
        elif m.gtype == "b":
            result[m.migration_key] = user_value.get_boolean()
        elif m.gtype == "s":
            result[m.migration_key] = user_value.get_string()
        elif m.gtype == "i":
            result[m.migration_key] = user_value.get_int32()
        elif m.gtype == "d":
            result[m.migration_key] = user_value.get_double()
        elif m.gtype == "as":
            result[m.migration_key] = list(user_value.unpack())
    return result


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def import_voice(gs: Gio.Settings, voice_data: dict, skip_defaults: bool = True) -> bool:
    """Import ACSS voice data to a Gio.Settings object. Returns True if any value was written."""

    migrated = False
    for acss_key, (gs_key, gs_type, default) in VOICE_MIGRATION_MAP.items():
        if acss_key not in voice_data:
            continue
        value = voice_data[acss_key]
        if skip_defaults:
            if gs_type == "b" and bool(value) == default:
                continue
            if gs_type == "i" and int(value) == default:
                continue
            if gs_type == "d" and float(value) == default:
                continue
        if gs_type == "b":
            gs.set_boolean(gs_key, bool(value))
        elif gs_type == "i":
            gs.set_int(gs_key, int(value))
        elif gs_type == "d":
            gs.set_double(gs_key, float(value))
        migrated = True

    family = voice_data.get("family", {})
    if isinstance(family, dict):
        for json_field, gs_key in VOICE_FAMILY_FIELDS.items():
            val = family.get(json_field)
            if val is not None and str(val):
                gs.set_string(gs_key, str(val))
                migrated = True

    return migrated


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def export_voice(gs: Gio.Settings) -> dict:
    """Export voice settings from a Gio.Settings to ACSS format."""

    voice_data: dict[str, Any] = {}
    for acss_key, (gs_key, gs_type, _default) in VOICE_MIGRATION_MAP.items():
        user_value = gs.get_user_value(gs_key)
        if user_value is None:
            continue
        if gs_type == "b":
            voice_data[acss_key] = user_value.get_boolean()
        elif gs_type == "i":
            voice_data[acss_key] = user_value.get_int32()
        elif gs_type == "d":
            voice_data[acss_key] = user_value.get_double()

    family_dict: dict[str, str] = {}
    for json_field, gs_key in VOICE_FAMILY_FIELDS.items():
        user_value = gs.get_user_value(gs_key)
        if user_value is not None:
            val = user_value.get_string()
            if val:
                family_dict[json_field] = val
    if family_dict:
        voice_data["family"] = family_dict

    return voice_data


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def import_synthesizer(gs: Gio.Settings, profile_prefs: dict) -> bool:
    """Import speechServerInfo to GSettings. Returns True if written."""

    speech_server_info = profile_prefs.get("speechServerInfo")
    if speech_server_info is None or len(speech_server_info) < 2:
        return False
    server_name = speech_server_info[0]
    synthesizer = speech_server_info[1]
    if server_name is None and synthesizer is None:
        return False
    gs.set_string("speech-server", server_name or "")
    gs.set_string("synthesizer", synthesizer or "")
    return True


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def export_synthesizer(gs: Gio.Settings) -> tuple[str, str] | None:
    """Export synthesizer from GSettings. Returns (display_name, module_id) or None."""

    synth_value = gs.get_user_value("synthesizer")
    if synth_value is None:
        return None
    synth = synth_value.get_string()
    server_value = gs.get_user_value("speech-server")
    server_name = server_value.get_string() if server_value is not None else ""
    if synth:
        return (server_name, synth)
    return ("", "")


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def import_pronunciations(gs: Gio.Settings, pronunciations_dict: dict) -> bool:
    """Import pronunciation dictionary to GSettings. Returns True if written.

    JSON format: {word: [word, replacement]} or {word: replacement}
    GSettings format: a{ss} {word: replacement}
    """

    converted: dict[str, str] = {}
    for key, value in pronunciations_dict.items():
        if isinstance(value, list) and len(value) >= 2:
            converted[key] = value[1]
        elif isinstance(value, list) and len(value) == 1:
            converted[key] = value[0]
        elif isinstance(value, str):
            converted[key] = value
    if not converted:
        return False

    gs.set_value("entries", GLib.Variant("a{ss}", converted))
    return True


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def export_pronunciations(gs: Gio.Settings) -> dict:
    """Export pronunciation dictionary from GSettings to JSON format.

    GSettings format: a{ss} {word: replacement}
    JSON format: {word: [word, replacement]}
    """

    user_value = gs.get_user_value("entries")
    if user_value is None:
        return {}
    entries = user_value.unpack()
    result: dict[str, list[str]] = {}
    for word, replacement in entries.items():
        result[word] = [word, replacement]
    return result


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def import_keybindings(gs: Gio.Settings, keybindings_dict: dict) -> bool:
    """Import keybinding overrides to GSettings. Returns True if written.

    JSON format: {command_name: [[keysym, mask, mods, clicks], ...]}
    GSettings format: a{saas} (same structure, all strings)
    """

    converted: dict[str, list[list[str]]] = {}
    for key, value in keybindings_dict.items():
        if key in KEYBINDINGS_METADATA_KEYS:
            continue
        if isinstance(value, list):
            bindings: list[list[str]] = [
                [str(v) for v in binding] for binding in value if isinstance(binding, list)
            ]
            converted[key] = bindings
    if not converted:
        return False

    gs.set_value("entries", GLib.Variant("a{saas}", converted))
    return True


# TODO - JD: Delete this in v52 (remove -i/--import-dir support).
def export_keybindings(gs: Gio.Settings) -> dict:
    """Export keybinding overrides from GSettings to JSON format."""

    user_value = gs.get_user_value("entries")
    if user_value is None:
        return {}
    return user_value.unpack()



# Orca
#
# Copyright 2024 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-return-statements

"""Utilities for accessible events."""

from __future__ import annotations

import enum
import threading
import time
from difflib import SequenceMatcher
from typing import TYPE_CHECKING

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import debug, focus_manager
from .ax_object import AXObject
from .ax_selection import AXSelection
from .ax_text import AXText
from .ax_utilities_collection import AXUtilitiesCollection
from .ax_utilities_component import AXUtilitiesComponent
from .ax_utilities_object import AXUtilitiesObject
from .ax_utilities_relation import AXUtilitiesRelation
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_state import AXUtilitiesState
from .ax_utilities_text import AXUtilitiesText
from .ax_value import AXValue

if TYPE_CHECKING:
    from typing import ClassVar

    from .input_event_manager import InputEventManager


class TextEventReason(enum.Enum):
    """Enum representing the reason for an object:text- event."""

    UNKNOWN = enum.auto()
    AUTO_DELETION = enum.auto()
    AUTO_INSERTION_PRESENTABLE = enum.auto()
    AUTO_INSERTION_UNPRESENTABLE = enum.auto()
    AUTO_SELECTION = enum.auto()
    AUTO_UNSELECTION = enum.auto()
    BACKSPACE = enum.auto()
    CHILDREN_CHANGE = enum.auto()
    CUT = enum.auto()
    DELETE = enum.auto()
    FOCUS_CHANGE = enum.auto()
    LIVE_REGION_UPDATE = enum.auto()
    MOUSE_MIDDLE_BUTTON = enum.auto()
    MOUSE_PRIMARY_BUTTON = enum.auto()
    NAVIGATION_BY_CHARACTER = enum.auto()
    NAVIGATION_BY_LINE = enum.auto()
    NAVIGATION_BY_PARAGRAPH = enum.auto()
    NAVIGATION_BY_PAGE = enum.auto()
    NAVIGATION_BY_WORD = enum.auto()
    NAVIGATION_TO_FILE_BOUNDARY = enum.auto()
    NAVIGATION_TO_LINE_BOUNDARY = enum.auto()
    PAGE_SWITCH = enum.auto()
    PASTE = enum.auto()
    REDO = enum.auto()
    SAY_ALL = enum.auto()
    SEARCH_PRESENTABLE = enum.auto()
    SEARCH_UNPRESENTABLE = enum.auto()
    SELECT_ALL = enum.auto()
    SELECTED_TEXT_DELETION = enum.auto()
    SELECTED_TEXT_INSERTION = enum.auto()
    SELECTED_TEXT_RESTORATION = enum.auto()
    SELECTION_BY_CHARACTER = enum.auto()
    SELECTION_BY_LINE = enum.auto()
    SELECTION_BY_PARAGRAPH = enum.auto()
    SELECTION_BY_PAGE = enum.auto()
    SELECTION_BY_WORD = enum.auto()
    SELECTION_TO_FILE_BOUNDARY = enum.auto()
    SELECTION_TO_LINE_BOUNDARY = enum.auto()
    SPIN_BUTTON_VALUE_CHANGE = enum.auto()
    SYSTEM_UPDATE = enum.auto()
    TYPING = enum.auto()
    TYPING_ECHOABLE = enum.auto()
    UI_UPDATE = enum.auto()
    UNDO = enum.auto()
    UNSPECIFIED_COMMAND = enum.auto()
    UNSPECIFIED_NAVIGATION = enum.auto()
    UNSPECIFIED_SELECTION = enum.auto()


class AXUtilitiesEvent:
    """Utilities for accessible events."""

    LAST_KNOWN_DESCRIPTION: ClassVar[dict[int, str]] = {}
    LAST_KNOWN_NAME: ClassVar[dict[int, str]] = {}

    LAST_KNOWN_CHECKED: ClassVar[dict[int, bool]] = {}
    LAST_KNOWN_EXPANDED: ClassVar[dict[int, bool]] = {}
    LAST_KNOWN_INDETERMINATE: ClassVar[dict[int, bool]] = {}
    LAST_KNOWN_INVALID_ENTRY: ClassVar[dict[int, bool]] = {}
    LAST_KNOWN_PRESSED: ClassVar[dict[int, bool]] = {}
    LAST_KNOWN_SELECTED: ClassVar[dict[int, bool]] = {}
    LAST_KNOWN_VALUE: ClassVar[dict[int, float]] = {}

    IGNORE_NAME_CHANGES_FOR: ClassVar[list[int]] = []

    TEXT_EVENT_REASON: ClassVar[dict[Atspi.Event, TextEventReason]] = {}

    _lock = threading.Lock()

    @staticmethod
    def _strings_are_redundant(str1: str | None, str2: str | None, threshold: float = 0.85) -> bool:
        """Returns True if str2 is redundant to str1 based on the similarity threshold."""

        if not (str1 and str2):
            return False

        similarity = round(SequenceMatcher(None, str1.lower(), str2.lower()).ratio(), 2)
        msg = (
            f"AXUtilitiesEvent: Similarity between '{str1}', '{str2}': {similarity} "
            f"(threshold: {threshold})"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return similarity >= threshold

    @staticmethod
    def _clear_stored_data() -> None:
        """Clears any data we have cached for objects"""

        while True:
            time.sleep(60)
            AXUtilitiesEvent._clear_all_dictionaries()

    @staticmethod
    def _clear_all_dictionaries(reason: str = "") -> None:
        msg = "AXUtilitiesEvent: Clearing local cache."
        if reason:
            msg += f" Reason: {reason}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        AXUtilitiesEvent.LAST_KNOWN_DESCRIPTION.clear()
        AXUtilitiesEvent.LAST_KNOWN_NAME.clear()
        AXUtilitiesEvent.LAST_KNOWN_CHECKED.clear()
        AXUtilitiesEvent.LAST_KNOWN_EXPANDED.clear()
        AXUtilitiesEvent.LAST_KNOWN_INDETERMINATE.clear()
        AXUtilitiesEvent.LAST_KNOWN_INVALID_ENTRY.clear()
        AXUtilitiesEvent.LAST_KNOWN_PRESSED.clear()
        AXUtilitiesEvent.LAST_KNOWN_SELECTED.clear()
        AXUtilitiesEvent.LAST_KNOWN_VALUE.clear()
        AXUtilitiesEvent.TEXT_EVENT_REASON.clear()
        AXUtilitiesEvent.IGNORE_NAME_CHANGES_FOR.clear()

    @staticmethod
    def clear_cache_now(reason: str = "") -> None:
        """Clears all cached information immediately."""

        AXUtilitiesEvent._clear_all_dictionaries(reason)

    @staticmethod
    def save_object_info_for_events(obj: Atspi.Accessible) -> None:
        """Saves properties, states, etc. of obj for later use in event processing."""

        if obj is None:
            return

        AXUtilitiesEvent.LAST_KNOWN_DESCRIPTION[hash(obj)] = AXObject.get_description(obj)
        AXUtilitiesEvent.LAST_KNOWN_NAME[hash(obj)] = AXObject.get_name(obj)
        AXUtilitiesEvent.LAST_KNOWN_CHECKED[hash(obj)] = AXUtilitiesState.is_checked(obj)
        AXUtilitiesEvent.LAST_KNOWN_EXPANDED[hash(obj)] = AXUtilitiesState.is_expanded(obj)
        AXUtilitiesEvent.LAST_KNOWN_INDETERMINATE[hash(obj)] = AXUtilitiesState.is_indeterminate(
            obj,
        )
        AXUtilitiesEvent.LAST_KNOWN_PRESSED[hash(obj)] = AXUtilitiesState.is_pressed(obj)
        AXUtilitiesEvent.LAST_KNOWN_SELECTED[hash(obj)] = AXUtilitiesState.is_selected(obj)

        window = focus_manager.get_manager().get_active_window()
        AXUtilitiesEvent.LAST_KNOWN_NAME[hash(window)] = AXObject.get_name(window)
        AXUtilitiesEvent.LAST_KNOWN_DESCRIPTION[hash(window)] = AXObject.get_description(window)

    @staticmethod
    def start_cache_clearing_thread() -> None:
        """Starts thread to periodically clear cached details."""

        thread = threading.Thread(target=AXUtilitiesEvent._clear_stored_data)
        thread.daemon = True
        thread.start()

    @staticmethod
    def get_text_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        reason = AXUtilitiesEvent.TEXT_EVENT_REASON.get(event)
        if reason is not None:
            tokens = ["AXUtilitiesEvent: Cached reason for", event, f": {reason}"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return reason

        reason = TextEventReason.UNKNOWN
        if event.type.startswith("object:text-changed:insert"):
            reason = AXUtilitiesEvent._get_text_insertion_event_reason(event)
        elif event.type.startswith("object:text-caret-moved"):
            reason = AXUtilitiesEvent._get_caret_moved_event_reason(event)
        elif event.type.startswith("object:text-changed:delete"):
            reason = AXUtilitiesEvent._get_text_deletion_event_reason(event)
        elif event.type.startswith("object:text-selection-changed"):
            reason = AXUtilitiesEvent._get_text_selection_changed_event_reason(event)
        else:
            raise ValueError(f"Unexpected event type: {event.type}")

        AXUtilitiesEvent.TEXT_EVENT_REASON[event] = reason
        tokens = ["AXUtilitiesEvent: Reason for", event, f": {reason}"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return reason

    @staticmethod
    def _get_selection_navigation_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the selection-granularity reason based on the last input event."""

        if mgr.last_event_was_line_navigation():
            reason = TextEventReason.SELECTION_BY_LINE
        elif mgr.last_event_was_word_navigation():
            reason = TextEventReason.SELECTION_BY_WORD
        elif mgr.last_event_was_character_navigation():
            reason = TextEventReason.SELECTION_BY_CHARACTER
        elif mgr.last_event_was_page_navigation():
            reason = TextEventReason.SELECTION_BY_PAGE
        elif mgr.last_event_was_line_boundary_navigation():
            reason = TextEventReason.SELECTION_TO_LINE_BOUNDARY
        elif mgr.last_event_was_file_boundary_navigation():
            reason = TextEventReason.SELECTION_TO_FILE_BOUNDARY
        else:
            reason = TextEventReason.UNSPECIFIED_SELECTION
        return reason

    @staticmethod
    def _get_caret_navigation_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the caret-navigation-granularity reason based on the last input event."""

        if mgr.last_event_was_line_navigation():
            reason = TextEventReason.NAVIGATION_BY_LINE
        elif mgr.last_event_was_word_navigation():
            reason = TextEventReason.NAVIGATION_BY_WORD
        elif mgr.last_event_was_character_navigation():
            reason = TextEventReason.NAVIGATION_BY_CHARACTER
        elif mgr.last_event_was_page_navigation():
            reason = TextEventReason.NAVIGATION_BY_PAGE
        elif mgr.last_event_was_line_boundary_navigation():
            reason = TextEventReason.NAVIGATION_TO_LINE_BOUNDARY
        elif mgr.last_event_was_file_boundary_navigation():
            reason = TextEventReason.NAVIGATION_TO_FILE_BOUNDARY
        else:
            reason = TextEventReason.UNSPECIFIED_NAVIGATION
        return reason

    @staticmethod
    def _get_editing_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the editing-operation reason based on the last input event."""

        if mgr.last_event_was_backspace():
            reason = TextEventReason.BACKSPACE
        elif mgr.last_event_was_delete():
            reason = TextEventReason.DELETE
        elif mgr.last_event_was_cut():
            reason = TextEventReason.CUT
        elif mgr.last_event_was_paste():
            reason = TextEventReason.PASTE
        elif mgr.last_event_was_undo():
            reason = TextEventReason.UNDO
        elif mgr.last_event_was_redo():
            reason = TextEventReason.REDO
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def _get_caret_moved_editable_reason(mgr: InputEventManager) -> TextEventReason:
        """Returns the reason for a caret-moved event in an editable or terminal."""

        reason = AXUtilitiesEvent._get_editing_reason(mgr)
        if reason != TextEventReason.UNKNOWN:
            return reason
        if mgr.last_event_was_page_switch():
            return TextEventReason.PAGE_SWITCH
        if mgr.last_event_was_command():
            return TextEventReason.UNSPECIFIED_COMMAND
        if mgr.last_event_was_printable_key():
            return TextEventReason.TYPING
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_caret_moved_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()

        obj = event.source
        mode, focus = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        if mode == focus_manager.SAY_ALL:
            reason = TextEventReason.SAY_ALL
        elif focus != obj and AXUtilitiesRole.is_text_input_search(focus):
            if mgr.last_event_was_backspace() or mgr.last_event_was_delete():
                reason = TextEventReason.SEARCH_UNPRESENTABLE
            else:
                reason = TextEventReason.SEARCH_PRESENTABLE
        elif mgr.last_event_was_caret_selection():
            reason = AXUtilitiesEvent._get_selection_navigation_reason(mgr)
        elif mgr.last_event_was_caret_navigation():
            reason = AXUtilitiesEvent._get_caret_navigation_reason(mgr)
        elif mgr.last_event_was_select_all():
            reason = TextEventReason.SELECT_ALL
        elif mgr.last_event_was_primary_click_or_release():
            reason = TextEventReason.MOUSE_PRIMARY_BUTTON
        elif AXUtilitiesState.is_editable(obj) or AXUtilitiesRole.is_terminal(obj):
            reason = AXUtilitiesEvent._get_caret_moved_editable_reason(mgr)
        elif mgr.last_event_was_tab_navigation():
            reason = TextEventReason.FOCUS_CHANGE
        elif AXUtilitiesObject.find_ancestor(obj, AXUtilitiesRole.children_are_presentational):
            reason = TextEventReason.UI_UPDATE
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def _get_deletion_editable_reason(
        event: Atspi.Event,
        mgr: InputEventManager,
    ) -> TextEventReason:
        """Returns the reason for a text-deletion event in an editable or terminal."""

        reason = AXUtilitiesEvent._get_editing_reason(mgr)
        if reason != TextEventReason.UNKNOWN:
            return reason

        obj = event.source
        if mgr.last_event_was_command():
            reason = TextEventReason.UNSPECIFIED_COMMAND
        elif mgr.last_event_was_printable_key():
            reason = TextEventReason.TYPING
        elif mgr.last_event_was_up_or_down() or mgr.last_event_was_page_up_or_page_down():
            if AXUtilitiesRole.is_spin_button(obj) or AXUtilitiesObject.find_ancestor(
                obj,
                AXUtilitiesRole.is_spin_button,
            ):
                reason = TextEventReason.SPIN_BUTTON_VALUE_CHANGE
            else:
                reason = TextEventReason.AUTO_DELETION
        else:
            selected_text, _start, _end = AXUtilitiesText.get_cached_selected_text(obj)
            if selected_text and event.any_data.strip() == selected_text.strip():
                reason = TextEventReason.SELECTED_TEXT_DELETION
        return reason

    @staticmethod
    def _get_text_deletion_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()

        obj = event.source
        if AXObject.get_role(obj) in AXUtilitiesRole.get_text_ui_roles():
            reason = TextEventReason.UI_UPDATE
        elif AXUtilitiesRole.is_live_region(obj):
            reason = TextEventReason.LIVE_REGION_UPDATE
        elif event.type.endswith("system"):
            reason = TextEventReason.SYSTEM_UPDATE
        elif mgr.last_event_was_page_switch():
            reason = TextEventReason.PAGE_SWITCH
        elif AXUtilitiesState.is_editable(obj) or AXUtilitiesRole.is_terminal(obj):
            reason = AXUtilitiesEvent._get_deletion_editable_reason(event, mgr)
        elif mgr.last_event_was_command():
            reason = TextEventReason.UNSPECIFIED_COMMAND
        elif "\ufffc" in event.any_data and not event.any_data.replace("\ufffc", ""):
            reason = TextEventReason.CHILDREN_CHANGE
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def _get_insertion_editable_reason(
        event: Atspi.Event,
        mgr: InputEventManager,
    ) -> TextEventReason:
        """Returns the reason for a text-insertion event in an editable or terminal."""

        obj = event.source
        selected_text, _start, _end = AXUtilitiesText.get_selected_text(obj)
        has_selected = bool(selected_text and event.any_data == selected_text)

        reason = AXUtilitiesEvent._get_insertion_editing_reason(mgr, has_selected)
        if reason != TextEventReason.UNKNOWN:
            return reason
        reason = AXUtilitiesEvent._get_insertion_typing_reason(event, mgr, obj, has_selected)
        if reason != TextEventReason.UNKNOWN:
            return reason
        if has_selected:
            return TextEventReason.SELECTED_TEXT_INSERTION
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_insertion_editing_reason(
        mgr: InputEventManager,
        has_selected: bool,
    ) -> TextEventReason:
        """Returns the editing-operation reason for a text insertion, if applicable."""

        if mgr.last_event_was_backspace():
            reason = TextEventReason.BACKSPACE
        elif mgr.last_event_was_delete():
            reason = TextEventReason.DELETE
        elif mgr.last_event_was_cut():
            reason = TextEventReason.CUT
        elif mgr.last_event_was_paste():
            reason = TextEventReason.PASTE
        elif mgr.last_event_was_undo():
            if has_selected:
                reason = TextEventReason.SELECTED_TEXT_RESTORATION
            else:
                reason = TextEventReason.UNDO
        elif mgr.last_event_was_redo():
            if has_selected:
                reason = TextEventReason.SELECTED_TEXT_RESTORATION
            else:
                reason = TextEventReason.REDO
        elif mgr.last_event_was_command():
            reason = TextEventReason.UNSPECIFIED_COMMAND
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def _get_insertion_text_key_reason(
        event: Atspi.Event,
        mgr: InputEventManager,
        obj: Atspi.Accessible,
        has_selected: bool,
    ) -> TextEventReason:
        """Returns the reason when the last event was a text-producing key."""

        if mgr.last_event_was_space() and not AXUtilitiesRole.is_password_text(obj):
            if event.any_data == "\n":
                reason = TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            else:
                reason = TextEventReason.TYPING
        elif mgr.last_event_was_tab() or mgr.last_event_was_return():
            if not event.any_data.strip():
                reason = TextEventReason.TYPING
            else:
                reason = AXUtilitiesEvent._get_insertion_fallback_reason(event, mgr)
        elif mgr.last_event_was_printable_key() or mgr.last_event_was_space():
            if has_selected:
                reason = TextEventReason.AUTO_INSERTION_PRESENTABLE
            else:
                reason = AXUtilitiesEvent._get_typing_echo_reason(obj)
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def _get_insertion_typing_reason(
        event: Atspi.Event,
        mgr: InputEventManager,
        obj: Atspi.Accessible,
        has_selected: bool,
    ) -> TextEventReason:
        """Returns the typing/auto-insertion reason for a text insertion."""

        reason = AXUtilitiesEvent._get_insertion_text_key_reason(event, mgr, obj, has_selected)
        if reason != TextEventReason.UNKNOWN:
            return reason

        if mgr.last_event_was_middle_click() or mgr.last_event_was_middle_release():
            reason = TextEventReason.MOUSE_MIDDLE_BUTTON
        elif mgr.last_event_was_up_or_down() or mgr.last_event_was_page_up_or_page_down():
            if AXUtilitiesRole.is_spin_button(obj) or AXUtilitiesObject.find_ancestor(
                obj,
                AXUtilitiesRole.is_spin_button,
            ):
                reason = TextEventReason.SPIN_BUTTON_VALUE_CHANGE
            else:
                reason = TextEventReason.AUTO_INSERTION_PRESENTABLE
        else:
            reason = AXUtilitiesEvent._get_insertion_fallback_reason(event, mgr)
        return reason

    @staticmethod
    def _get_typing_echo_reason(obj: Atspi.Accessible) -> TextEventReason:
        """Returns TYPING_ECHOABLE if echo is enabled, otherwise TYPING."""

        from . import typing_echo_presenter  # pylint: disable=import-outside-toplevel

        presenter = typing_echo_presenter.get_presenter()
        if AXUtilitiesRole.is_password_text(obj):
            echo = presenter.get_key_echo_enabled()
        else:
            echo = presenter.get_character_echo_enabled()
        if echo:
            return TextEventReason.TYPING_ECHOABLE
        return TextEventReason.TYPING

    @staticmethod
    def _get_insertion_fallback_reason(
        event: Atspi.Event,
        mgr: InputEventManager,
    ) -> TextEventReason:
        """Returns the reason for unclassified text insertions."""

        if mgr.last_event_was_tab() and event.any_data != "\t":
            return TextEventReason.AUTO_INSERTION_PRESENTABLE
        if mgr.last_event_was_return() and event.any_data != "\n":
            if AXUtilitiesState.is_single_line(event.source):
                return TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            return TextEventReason.AUTO_INSERTION_PRESENTABLE
        if len(event.any_data) == 1:
            return TextEventReason.UNKNOWN
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_text_insertion_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()

        obj = event.source
        if AXObject.get_role(obj) in AXUtilitiesRole.get_text_ui_roles():
            reason = TextEventReason.UI_UPDATE
        elif AXUtilitiesRole.is_live_region(obj):
            reason = TextEventReason.LIVE_REGION_UPDATE
        elif event.type.endswith("system"):
            reason = TextEventReason.SYSTEM_UPDATE
        elif mgr.last_event_was_page_switch():
            reason = TextEventReason.PAGE_SWITCH
        elif AXUtilitiesState.is_editable(obj) or AXUtilitiesRole.is_terminal(obj):
            reason = AXUtilitiesEvent._get_insertion_editable_reason(event, mgr)
        elif mgr.last_event_was_command():
            reason = TextEventReason.UNSPECIFIED_COMMAND
        elif "\ufffc" in event.any_data and not event.any_data.replace("\ufffc", ""):
            reason = TextEventReason.CHILDREN_CHANGE
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def _get_selection_changed_editable_reason(
        obj: Atspi.Accessible,
        mgr: InputEventManager,
    ) -> TextEventReason:
        """Returns the reason for a text-selection-changed event in an editable or terminal."""

        reason = AXUtilitiesEvent._get_editing_reason(mgr)
        if reason != TextEventReason.UNKNOWN:
            return reason
        if mgr.last_event_was_page_switch():
            return TextEventReason.PAGE_SWITCH
        if mgr.last_event_was_command():
            return TextEventReason.UNSPECIFIED_COMMAND
        if mgr.last_event_was_printable_key():
            cached_text, old_start, old_end = AXUtilitiesText.get_cached_selected_text(obj)
            if not cached_text:
                _current, new_start, new_end = AXUtilitiesText.get_selected_text(obj)
                if new_start < new_end:
                    return TextEventReason.AUTO_SELECTION
            else:
                _current, new_start, new_end = AXUtilitiesText.get_selected_text(obj)
                if new_start == old_start + 1 and new_end == old_end:
                    return TextEventReason.AUTO_UNSELECTION
            return TextEventReason.TYPING
        if mgr.last_event_was_up_or_down() or mgr.last_event_was_page_up_or_page_down():
            if AXUtilitiesRole.is_spin_button(obj) or AXUtilitiesObject.find_ancestor(
                obj,
                AXUtilitiesRole.is_spin_button,
            ):
                return TextEventReason.SPIN_BUTTON_VALUE_CHANGE
        return TextEventReason.UNKNOWN

    @staticmethod
    def _get_text_selection_changed_event_reason(event: Atspi.Event) -> TextEventReason:
        """Returns the TextEventReason for the given event."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()

        obj = event.source
        focus = focus_manager.get_manager().get_locus_of_focus()
        if focus != obj and AXUtilitiesRole.is_text_input_search(focus):
            if mgr.last_event_was_backspace() or mgr.last_event_was_delete():
                reason = TextEventReason.SEARCH_UNPRESENTABLE
            else:
                reason = TextEventReason.SEARCH_PRESENTABLE
        elif mgr.last_event_was_caret_selection():
            reason = AXUtilitiesEvent._get_selection_navigation_reason(mgr)
        elif mgr.last_event_was_caret_navigation():
            reason = AXUtilitiesEvent._get_caret_navigation_reason(mgr)
        elif mgr.last_event_was_select_all():
            reason = TextEventReason.SELECT_ALL
        elif mgr.last_event_was_primary_click_or_release():
            reason = TextEventReason.MOUSE_PRIMARY_BUTTON
        elif AXUtilitiesState.is_editable(obj) or AXUtilitiesRole.is_terminal(obj):
            reason = AXUtilitiesEvent._get_selection_changed_editable_reason(obj, mgr)
        else:
            reason = TextEventReason.UNKNOWN
        return reason

    @staticmethod
    def is_presentable_active_descendant_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an active-descendant change."""

        if not event.any_data:
            msg = "AXUtilitiesEvent: No any_data."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not (
            AXUtilitiesState.is_focused(event.source) or AXUtilitiesState.is_focused(event.any_data)
        ):
            msg = "AXUtilitiesEvent: Neither source nor child have focused state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if AXUtilitiesRole.is_table_cell(focus):
            table = AXUtilitiesObject.find_ancestor(focus, AXUtilitiesRole.is_tree_or_tree_table)
            if table is not None and table != event.source:
                msg = "AXUtilitiesEvent: Event is from a different tree or tree table."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_checked_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a checked-state change."""

        old_state = AXUtilitiesEvent.LAST_KNOWN_CHECKED.get(hash(event.source))
        new_state = AXUtilitiesState.is_checked(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_CHECKED[hash(event.source)] = new_state
        focus = focus_manager.get_manager().get_locus_of_focus()
        if event.source != focus:
            if not AXUtilitiesObject.is_ancestor(event.source, focus):
                msg = "AXUtilitiesEvent: The source is not the locus of focus or its descendant."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False
            if not (AXUtilitiesRole.is_list_item(focus) or AXUtilitiesRole.is_tree_item(focus)):
                msg = "AXUtilitiesEvent: The source descends from non-interactive-item focus."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        mgr = input_event_manager.get_manager()

        # Radio buttons normally change their state when you arrow to them, so we handle the
        # announcement of their state changes in the focus handling code.
        if AXUtilitiesRole.is_radio_button(event.source) and not mgr.last_event_was_space():
            msg = "AXUtilitiesEvent: Only presentable for this role if toggled by user."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_changed_description_to_present(event: Atspi.Event) -> bool:
        """Returns True if the description data warrants further presentability checks."""

        if not isinstance(event.any_data, str):
            msg = "AXUtilitiesEvent: The any_data is not a string."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        old_description = AXUtilitiesEvent.LAST_KNOWN_DESCRIPTION.get(hash(event.source))
        new_description = event.any_data
        if old_description == new_description:
            msg = "AXUtilitiesEvent: The new description matches the old description."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesEvent._strings_are_redundant(old_description, new_description):
            msg = (
                f"AXUtilitiesEvent: The new description ('{new_description}') "
                f"is too similar to the old description ('{old_description}')."
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_DESCRIPTION[hash(event.source)] = new_description
        if not new_description:
            msg = "AXUtilitiesEvent: The description is empty."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        return True

    @staticmethod
    def is_presentable_description_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a description change."""

        if not AXUtilitiesEvent._is_changed_description_to_present(event):
            return False

        if not AXUtilitiesState.is_showing(event.source):
            msg = "AXUtilitiesEvent: The event source is not showing."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if event.source != focus and not AXUtilitiesObject.is_ancestor(focus, event.source):
            msg = "AXUtilitiesEvent: The event is not from the locus of focus or ancestor."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_expanded_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an expanded-state change."""

        old_state = AXUtilitiesEvent.LAST_KNOWN_EXPANDED.get(hash(event.source))
        new_state = AXUtilitiesState.is_expanded(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_EXPANDED[hash(event.source)] = new_state
        focus = focus_manager.get_manager().get_locus_of_focus()
        if event.source == focus:
            msg = "AXUtilitiesEvent: Event is presentable, from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not event.detail1 and not AXUtilitiesObject.is_ancestor(focus, event.source):
            msg = "AXUtilitiesEvent: Event is not from the locus of focus or ancestor."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_table_row(event.source) or AXUtilitiesRole.is_list_box(event.source):
            msg = "AXUtilitiesEvent: Event is presentable based on role."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilitiesRole.is_combo_box(event.source) or AXUtilitiesRole.is_button(event.source):
            if not AXUtilitiesState.is_focused(event.source):
                msg = "AXUtilitiesEvent: Only presentable for this role if focused."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_indeterminate_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an indeterminate-state change."""

        old_state = AXUtilitiesEvent.LAST_KNOWN_INDETERMINATE.get(hash(event.source))
        new_state = AXUtilitiesState.is_indeterminate(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_INDETERMINATE[hash(event.source)] = new_state

        # If this state is cleared, the new state will become checked or unchecked
        # and we should get object:state-changed:checked events for those cases.
        if not new_state:
            msg = "AXUtilitiesEvent: The new state should be presented as a checked-change."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_invalid_entry_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as an invalid-entry-state change."""

        old_state = AXUtilitiesEvent.LAST_KNOWN_INVALID_ENTRY.get(hash(event.source))
        new_state = AXUtilitiesState.is_invalid_entry(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_INVALID_ENTRY[hash(event.source)] = new_state
        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_presentable_frame_name_change(event: Atspi.Event) -> bool:
        """Returns True if a frame's name-change event should be presented."""

        if event.source != focus_manager.get_manager().get_active_window():
            msg = "AXUtilitiesEvent: Event is for frame other than the active window."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        # Example: Typing the subject in an email client causing the window name to change.
        focus = focus_manager.get_manager().get_locus_of_focus()
        if (
            AXUtilitiesState.is_editable(focus)
            and AXText.get_character_count(focus)
            and AXText.get_all_text(focus) in event.any_data
        ):
            msg = "AXUtilitiesEvent: Event is redundant notification for the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_changed_name_to_present(event: Atspi.Event) -> bool:
        """Returns True if the name data warrants further presentability checks."""

        if hash(event.source) in AXUtilitiesEvent.IGNORE_NAME_CHANGES_FOR:
            msg = "AXUtilitiesEvent: Ignoring name change for this source."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not isinstance(event.any_data, str):
            msg = "AXUtilitiesEvent: The any_data is not a string."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        old_name = AXUtilitiesEvent.LAST_KNOWN_NAME.get(hash(event.source))
        new_name = event.any_data
        if old_name == new_name:
            msg = "AXUtilitiesEvent: The new name matches the old name."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesEvent._strings_are_redundant(old_name, new_name):
            msg = (
                f"AXUtilitiesEvent: The new name ('{new_name}') "
                f"is too similar to the old name ('{old_name}')."
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_NAME[hash(event.source)] = new_name
        if not new_name:
            msg = "AXUtilitiesEvent: The name is empty."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        return True

    @staticmethod
    def is_presentable_name_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a name change."""

        if not AXUtilitiesEvent._is_changed_name_to_present(event):
            return False

        if not AXUtilitiesState.is_showing(event.source):
            msg = "AXUtilitiesEvent: The event source is not showing."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if event.source != focus_manager.get_manager().get_locus_of_focus():
                return False

        if AXUtilitiesRole.is_frame(event.source):
            return AXUtilitiesEvent._is_presentable_frame_name_change(event)

        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_list_item(event.source):
            # Inspired by Firefox's downloads list in which the list item's name changes to show
            # download progress. It's super chatty and stomps on the progress bar announcements.
            if AXObject.supports_collection(event.source):
                has_progress_bar = (
                    AXUtilitiesCollection.find_first_with_role(
                        event.source, [Atspi.Role.PROGRESS_BAR]
                    )
                    is not None
                )
            else:
                has_progress_bar = (
                    AXUtilitiesObject.find_descendant(event.source, AXUtilitiesRole.is_progress_bar)
                    is not None
                )
            if has_progress_bar:
                msg = "AXUtilitiesEvent: The list item contains a progress bar."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                AXUtilitiesEvent.IGNORE_NAME_CHANGES_FOR.append(hash(event.source))
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_pressed_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a pressed-state change."""

        old_state = AXUtilitiesEvent.LAST_KNOWN_PRESSED.get(hash(event.source))
        new_state = AXUtilitiesState.is_pressed(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_PRESSED[hash(event.source)] = new_state
        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_selected_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a selected-state change."""

        old_state = AXUtilitiesEvent.LAST_KNOWN_SELECTED.get(hash(event.source))
        new_state = AXUtilitiesState.is_selected(event.source)
        if old_state == new_state:
            msg = "AXUtilitiesEvent: The new state matches the old state."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        AXUtilitiesEvent.LAST_KNOWN_SELECTED[hash(event.source)] = new_state
        if event.source != focus_manager.get_manager().get_locus_of_focus():
            msg = "AXUtilitiesEvent: The event is not from the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_selection_change(event: Atspi.Event) -> bool:
        """Returns True if this object:selection-changed event should be presented."""

        from . import input_event_manager  # pylint: disable=import-outside-toplevel

        if input_event_manager.get_manager().last_event_was_space():
            msg = "AXUtilitiesEvent: Selection toggled via space; handled elsewhere."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesState.manages_descendants(event.source):
            msg = "AXUtilitiesEvent: Source manages descendants; handled elsewhere."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_menu(event.source):
            if AXUtilitiesState.is_showing(event.source) and AXUtilitiesState.is_visible(
                event.source
            ):
                msg = "AXUtilitiesEvent: Event is presentable: Source is a menu."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            # This is a sad workaround for GTK2 menu items with submenus losing their showing state
            # when submenu children become selected.
            child = AXSelection.get_selected_child(event.source, 0)
            if (
                child is not None
                and AXUtilitiesState.is_showing(child)
                and AXUtilitiesState.is_visible(child)
            ):
                tokens = [
                    "AXUtilitiesEvent: Event is presentable: Selected child",
                    child,
                    "is showing and visible",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return True

            tokens = ["AXUtilitiesEvent: Menu lacks showing + visible:", event.source]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        if AXUtilitiesRole.is_combo_box(event.source) and not AXUtilitiesState.is_expanded(
            event.source
        ):
            text_input = AXUtilitiesObject.find_descendant(
                event.source, AXUtilitiesRole.is_text_input
            )
            if text_input is not None and AXUtilitiesState.is_focused(text_input):
                msg = "AXUtilitiesEvent: Combo box text input is focused; not a user selection."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

        focus = focus_manager.get_manager().get_locus_of_focus()

        mgr = input_event_manager.get_manager()
        if (
            AXUtilitiesState.supports_autocompletion(focus)
            and not mgr.last_event_was_up_or_down()
            and not mgr.last_event_was_page_up_or_page_down()
        ):
            if AXUtilitiesRelation.object_is_controlled_by(event.source, focus):
                msg = "AXUtilitiesEvent: Source is autocomplete for focused widget; not presenting."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False
        if event.source != focus and not (
            AXUtilitiesState.is_showing(event.source) and AXUtilitiesState.is_visible(event.source)
        ):
            combobox = AXUtilitiesObject.find_ancestor(event.source, AXUtilitiesRole.is_combo_box)
            if combobox != focus and event.source != AXObject.get_parent(focus):
                tokens = ["AXUtilitiesEvent: Source lacks showing + visible:", event.source]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        active_window = focus_manager.get_manager().get_active_window()
        if active_window and not AXUtilitiesObject.is_ancestor(event.source, active_window):
            tokens = ["AXUtilitiesEvent:", event.source, "is not inside", active_window]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        msg = "AXUtilitiesEvent: Event is presentable: Not ruled out by any checks."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_value_change(event: Atspi.Event) -> bool:
        """Returns True if this event should be presented as a value change."""

        old_value = AXUtilitiesEvent.LAST_KNOWN_VALUE.get(hash(event.source))
        new_value = AXValue.get_current_value(event.source)
        AXUtilitiesEvent.LAST_KNOWN_VALUE[hash(event.source)] = new_value
        if old_value == new_value:
            msg = "AXUtilitiesEvent: The new value matches the old value."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilitiesRole.is_progress_bar(event.source):
            if AXUtilitiesComponent.has_no_size(event.source):
                tokens = ["AXUtilitiesEvent:", event.source, "has no size."]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False
            if AXUtilitiesObject.find_ancestor(event.source, AXUtilitiesRole.is_status_bar):
                tokens = ["AXUtilitiesEvent:", event.source, "is in a status bar."]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def _is_presentable_text_event(event: Atspi.Event) -> bool:
        """Returns True if this text event should be presented."""

        if not (
            AXUtilitiesState.is_editable(event.source) or AXUtilitiesRole.is_terminal(event.source)
        ):
            msg = "AXUtilitiesEvent: The source is neither editable nor a terminal."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if focus != event.source and not AXUtilitiesState.is_focused(event.source):
            msg = "AXUtilitiesEvent: The source is neither focused, nor the locus of focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)

            # This can happen in web content where the focus is a contenteditable element and a
            # new child element is created for new or changed text.
            if AXUtilitiesObject.is_ancestor(event.source, focus):
                msg = "AXUtilitiesEvent: The locus of focus is an ancestor of the source."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            return False

        msg = "AXUtilitiesEvent: Event is presentable."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    @staticmethod
    def is_presentable_text_attributes_change(event: Atspi.Event) -> bool:
        """Returns True if this text-attributes-change event should be presented."""

        return AXUtilitiesEvent._is_presentable_text_event(event)

    @staticmethod
    def is_presentable_text_deletion(event: Atspi.Event) -> bool:
        """Returns True if this text-deletion event should be presented."""

        return AXUtilitiesEvent._is_presentable_text_event(event)

    @staticmethod
    def is_presentable_text_insertion(event: Atspi.Event) -> bool:
        """Returns True if this text-insertion event should be presented."""

        return AXUtilitiesEvent._is_presentable_text_event(event)


AXUtilitiesEvent.start_cache_clearing_thread()
